@@ -1,5 +1,5 @@
 /*
- * Copyright 2009 - 2010 by Ping Cheng, Wacom. <pingc@wacom.com>
+ * Copyright 2009 - 2013 by Ping Cheng, Wacom. <pingc@wacom.com>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -35,12 +35,17 @@ static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
 	int match = 0;
 	InputInfoPtr pDevices = xf86FirstLocalDevice();
 
-	for (; pDevices != NULL; pDevices = pDevices->next)
+	for (; !match && pDevices != NULL; pDevices = pDevices->next)
 	{
 		char* device = xf86CheckStrOption(pDevices->options, "Device", NULL);
 
 		/* device can be NULL on some distros */
-		if (!device || !strstr(pDevices->drv->driverName, "wacom"))
+		if (!device)
+			continue;
+
+		free(device);
+
+		if (!strstr(pDevices->drv->driverName, "wacom"))
 			continue;
 
 		if (pInfo != pDevices)
@@ -55,11 +60,10 @@ static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
 				/* only add the new tool if the matching major/minor
 				* was from the same source */
 				if (strcmp(fsource, psource))
-				{
 					match = 1;
-					break;
-				}
 			}
+			free(fsource);
+			free(psource);
 		}
 	}
 	if (match)
@@ -112,20 +116,21 @@ int wcmIsDuplicate(const char* device, InputInfoPtr pInfo)
 		isInUse = 4;
 	}
 ret:
+	free(lsource);
 	return isInUse;
 }
 
 static struct
 {
 	const char* type;
-	__u16 tool[3]; /* tool array is terminated by 0 */
+	__u16 tool[5]; /* tool array is terminated by 0 */
 } wcmType [] =
 {
 	{ "stylus", { BTN_TOOL_PEN,       0                  } },
 	{ "eraser", { BTN_TOOL_RUBBER,    0                  } },
 	{ "cursor", { BTN_TOOL_MOUSE,     0                  } },
 	{ "touch",  { BTN_TOOL_DOUBLETAP, BTN_TOOL_FINGER, 0 } },
-	{ "pad",    { BTN_FORWARD,        BTN_0,           0 } }
+	{ "pad",    { BTN_FORWARD,        BTN_0,           KEY_CONTROLPANEL, KEY_ONSCREEN_KEYBOARD, 0 } }
 };
 
 /* validate tool type for device/product */
@@ -134,7 +139,7 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 	int j, k, ret = FALSE;
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
-	char* dsource = xf86CheckStrOption(pInfo->options, "_source", NULL);
+	char* dsource;
 
 	if (!type)
 	{
@@ -142,6 +147,8 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 		return FALSE;
 	}
 
+	dsource = xf86CheckStrOption(pInfo->options, "_source", NULL);
+
 	/* walkthrough all types */
 	for (j = 0; j < ARRAY_SIZE(wcmType); j++)
 	{
@@ -175,6 +182,7 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 		xf86Msg(X_ERROR, "%s: Invalid type '%s' for this device.\n",
 			pInfo->name, type);
 
+	free(dsource);
 	return ret;
 }
 
@@ -191,9 +199,26 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 	{
 		case 0xF8:  /* Cintiq 24HDT */
 		case 0xF4:  /* Cintiq 24HD */
-			TabletSetFeature(priv->common, WCM_DUALRING | WCM_LCD);
+			TabletSetFeature(priv->common, WCM_DUALRING);
 			/* fall through */
 
+		case 0x34D: /* MobileStudio Pro 13 */
+		case 0x34E: /* MobileStudio Pro 16 */
+		case 0x398: /* MobileStudio Pro 13 */
+		case 0x399: /* MobileStudio Pro 16 */
+		case 0x3AA: /* MobileStudio Pro 16 */
+			TabletSetFeature(priv->common, WCM_LCD);
+			/* fall through */
+
+		case 0x357: /* Intuos Pro 2 M */
+		case 0x358: /* Intuos Pro 2 L */
+		case 0x360: /* Intuos Pro 2 M (Bluetooth) */
+		case 0x36a: /* Intuos Pro 2 L (Bluetooth) */
+		case 0x392: /* Intuos Pro 2 S */
+		case 0x393: /* Intuos Pro 2 S (Bluetooth) */
+		case 0x314: /* Intuos Pro S */
+		case 0x315: /* Intuos Pro M */
+		case 0x317: /* Intuos Pro L */
 		case 0x26:  /* I5 */
 		case 0x27:  /* I5 */
 		case 0x28:  /* I5 */
@@ -246,6 +271,9 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 		case 0xC7:  /* DTU1931 */
 		case 0xCE:  /* DTU2231 */
 		case 0xF0:  /* DTU1631 */
+
+		/* Wacom One display tablet */
+		case 0x3A6: /* DTC133 */
 			TabletSetFeature(priv->common, WCM_LCD);
 			break;
 
@@ -255,6 +283,7 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 		case 0xC6:  /* CintiqV5 */
 		case 0xCC:  /* CinitqV5 */
 		case 0xFA:  /* Cintiq 22HD */
+		case 0x5B:  /* Cintiq 22HDT Pen */
 			TabletSetFeature(priv->common, WCM_LCD);
 			/* fall through */
 		case 0xB0:  /* I3 */
@@ -269,26 +298,90 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 
 		case 0x100: /* TPC with MT */
 		case 0x101: /* TPC with MT */
+		case 0x10D: /* TPC with MT */
+		case 0x116: /* TPC with 1FGT */
+		case 0x12C: /* TPC */
 		case 0x4001: /* TPC with MT */
+		case 0x4004: /* TPC with MT (no pen on Motion) */
+		case 0x5000: /* TPC with MT */
+		case 0x5002: /* TPC with MT */
 		case 0xE2: /* TPC with 2FGT */
 		case 0xE3: /* TPC with 2FGT */
 		case 0xE5: /* TPC with MT */
 		case 0xE6: /* TPC with 2FGT */
 		case 0x93: /* TPC with 1FGT */
 		case 0x9A: /* TPC with 1FGT */
+		case 0xEC: /* TPC with 1FGT */
 		case 0xED: /* TPC with 1FGT */
 		case 0x90: /* TPC */
 		case 0x97: /* TPC */
+		case 0x9F: /* TPC */
 		case 0xEF: /* TPC */
 			TabletSetFeature(priv->common, WCM_TPC);
 			break;
 
-		case 0x9F:
+		case 0x304:/* Cintiq 13HD */
+		case 0x307:/* Cintiq Companion Hybrid */
+		case 0x30A:/* Cintiq Companion */
+		case 0x325:/* Cintiq Companion 2 */
+		case 0x32A:/* Cintiq 27QHD */
+		case 0x32B:/* Cintiq 27QHDT Pen */
+		case 0x333:/* Cintiq 13HDT Pen */
+		case 0x34F:/* Cintiq Pro 13 FHD */
+		case 0x350:/* Cintiq Pro 16 UHD */
+		case 0x351:/* Cintiq Pro 24 */
+		case 0x352:/* Cintiq Pro 32 */
+		case 0x37C:/* Cintiq Pro 24 Pen-Only */
+		case 0x390:/* Cintiq 16 */
+		case 0x391:/* Cintiq 22 */
+		case 0x396:/* DTK-1660E */
+		case 0x3AE:/* Cintiq 16 */
+		case 0x3B0:/* DTK-1660E */
+			TabletSetFeature(priv->common, WCM_ROTATION);
+			/* fall-through */
+
 		case 0xF6: /* Cintiq 24HDT Touch */
+		case 0x57: /* DTK2241 */
+		case 0x59: /* DTH2242 Pen */
+		case 0x5D: /* DTH2242 Touch */
+		case 0x5E: /* Cintiq 22HDT Touch */
+		case 0x309:/* Cintiq Companion Hybrid Touch */
+		case 0x30C:/* Cintiq Companion Touch */
+		case 0x326:/* Cintiq Companion 2 Touch */
+		case 0x32C:/* Cintiq 27QHDT Touch */
+		case 0x32F:/* DTU-1031X */
+		case 0x335:/* Cintiq 13HDT Touch */
+		case 0x336:/* DTU-1141 */
+		case 0x343:/* DTK-1651 */
+		case 0x34A:/* MobileStudio Pro 13 Touch */
+		case 0x34B:/* MobileStudio Pro 16 Touch */
+		case 0x353:/* Cintiq Pro 13 FHD Touch */
+		case 0x354:/* Cintiq Pro 13 UHD Touch */
+		case 0x355:/* Cintiq Pro 24 Touch */
+		case 0x356:/* Cintiq Pro 32 Touch */
+		case 0x359:/* DTU-1141B */
+		case 0x35A:/* DTH-1152*/
+		case 0x368:/* DTH-1152 Touch */
+		case 0x382:/* DTK-2451 */
+		case 0x37D:/* DTH-2452 */
+		case 0x37E:/* DTH-2452 Touch */
+		case 0x39A:/* MobileStudio Pro 13 Touch */
+		case 0x39B:/* MobileStudio Pro 16 Touch */
+		case 0x3AC:/* MobileStudio Pro 16 Touch */
 			TabletSetFeature(priv->common, WCM_LCD);
 			break;
 	}
 
+#ifdef INPUT_PROP_DIRECT
+	{
+		int rc;
+		unsigned long prop[NBITS(INPUT_PROP_MAX)] = {0};
+
+		rc = ioctl(pInfo->fd, EVIOCGPROP(sizeof(prop)), prop);
+		if (rc >= 0 && ISBITSET(prop, INPUT_PROP_DIRECT))
+			TabletSetFeature(priv->common, WCM_LCD);
+	}
+#endif
 	if (ISBITSET(common->wcmKeys, BTN_TOOL_PEN))
 		TabletSetFeature(priv->common, WCM_PEN);
 
@@ -329,8 +422,8 @@ input_option_new(InputOption *list, char *key, char *value)
 	InputOption *new;
 
 	new = calloc(1, sizeof(InputOption));
-	new->key = key;
-	new->value = value;
+	new->key = strdup(key);
+	new->value = strdup(value);
 	new->next = list;
 	return new;
 }
@@ -366,7 +459,7 @@ static InputOption *wcmOptionDupConvert(InputInfoPtr pInfo, const char* basename
 	WacomToolPtr ser = common->serials;
 	InputOption *iopts = NULL;
 	char *name;
-	pointer options;
+	pointer options, o;
 	int rc;
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
@@ -404,13 +497,15 @@ static InputOption *wcmOptionDupConvert(InputInfoPtr pInfo, const char* basename
 
 	free(name);
 
-	while(options)
+	o = options;
+	while(o)
 	{
 		iopts = input_option_new(iopts,
-					 xf86OptionName(options),
-					 xf86OptionValue(options));
-		options = xf86NextOption(options);
+					 xf86OptionName(o),
+					 xf86OptionValue(o));
+		o = xf86NextOption(o);
 	}
+	xf86OptionListFree(options);
 	return iopts;
 }
 
@@ -426,10 +521,12 @@ static InputAttributes* wcmDuplicateAttributes(InputInfoPtr pInfo,
 {
 	int rc;
 	InputAttributes *attr;
+	char *product;
+
 	attr = DuplicateInputAttributes(pInfo->attrs);
-	rc = asprintf(&attr->product, "%s %s", attr->product, type);
-	if (rc == -1)
-		attr->product = NULL;
+	rc = asprintf(&product, "%s %s", attr->product, type);
+	free(attr->product);
+	attr->product = (rc != -1) ? product : NULL;
 	return attr;
 }
 #endif
@@ -460,11 +557,19 @@ wcmHotplugDevice(ClientPtr client, pointer closure )
 	WacomHotplugInfo *hotplug_info = closure;
 	DeviceIntPtr dev; /* dummy */
 
+#if HAVE_THREADED_INPUT
+	input_lock();
+#endif
+
 	NewInputDeviceRequest(hotplug_info->input_options,
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 9
 			      hotplug_info->attrs,
 #endif
 			      &dev);
+#if HAVE_THREADED_INPUT
+	input_unlock();
+#endif
+
 	input_option_free_list(&hotplug_info->input_options);
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 11
@@ -483,7 +588,7 @@ wcmHotplugDevice(ClientPtr client, pointer closure )
  *
  * Note that we don't actually hotplug the device here. We store the
  * information needed to hotplug the device later and then queue the
- * hotplug. The server will come back and call the @wcmHotplugDevice
+ * hotplug. The server will come back and call the @ref wcmHotplugDevice
  * later.
  *
  * @param pInfo The parent device
@@ -574,19 +679,20 @@ void wcmHotplugOthers(InputInfoPtr pInfo, const char *basename)
  * This changes the source to _driver/wacom, all auto-hotplugged devices
  * will have the same source.
  */
-int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
+int wcmNeedAutoHotplug(InputInfoPtr pInfo, char **type)
 {
 	char *source = xf86CheckStrOption(pInfo->options, "_source", NULL);
 	int i;
+	int rc = 0;
 
 	if (*type) /* type specified, don't hotplug */
-		return 0;
+		goto out;
 
 	if (!source) /* xorg.conf device, don't auto-pick type */
-		return 0;
+		goto out;
 
 	if (source && strcmp(source, "server/hal") && strcmp(source, "server/udev"))
-		return 0;
+		goto out;
 
 	/* no type specified, so we need to pick the first one applicable
 	 * for our device */
@@ -594,13 +700,14 @@ int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
 	{
 		if (wcmIsAValidType(pInfo, wcmType[i].type))
 		{
+			free(*type);
 			*type = strdup(wcmType[i].type);
 			break;
 		}
 	}
 
 	if (!*type)
-		return 0;
+		goto out;
 
 	xf86Msg(X_INFO, "%s: type not specified, assuming '%s'.\n", pInfo->name, *type);
 	xf86Msg(X_INFO, "%s: other types will be automatically added.\n", pInfo->name);
@@ -609,7 +716,11 @@ int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
 	pInfo->options = xf86AddNewOption(pInfo->options, "Type", *type);
 	pInfo->options = xf86ReplaceStrOption(pInfo->options, "_source", "_driver/wacom");
 
-	return 1;
+	rc = 1;
+
+	free(source);
+out:
+	return rc;
 }
 
 int wcmParseSerials (InputInfoPtr pInfo)
@@ -677,7 +788,7 @@ int wcmParseSerials (InputInfoPtr pInfo)
 					pInfo->name, serial, name);
 				ser->name = strdup(name);
 			}
-			else ser->name = ""; /*no name yet*/
+			else ser->name = strdup(""); /*no name yet*/
 
 			if (common->serials == NULL)
 				common->serials = ser;
@@ -706,7 +817,7 @@ int wcmParseSerials (InputInfoPtr pInfo)
  * hotplugging, False if the device is a depent or xorg.conf device.
  * @param is_hotplugged True if the device is a dependent device, FALSE
  * otherwise.
- * @retvalue True on success or False otherwise.
+ * @retval True on success or False otherwise.
  */
 Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 			    Bool is_dependent)
@@ -736,6 +847,8 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 		 */
 	}
 
+	free(s);
+
 	/* Pad is always in absolute mode.
 	 * The pad also defaults to wheel scrolling, unlike the pens
 	 * (interesting effects happen on ArtPen and others with build-in
@@ -772,6 +885,7 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 					" device\n", pInfo->name);
 		else
 			wcmRotateTablet(pInfo, rotation);
+		free(s);
 	}
 
 	common->wcmRawSample = xf86SetIntOption(pInfo->options, "RawSample",
@@ -793,10 +907,6 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 			common->wcmSuppress = DEFAULT_SUPPRESS;
 	}
 
-	if (xf86SetBoolOption(pInfo->options, "Tilt",
-			(common->wcmFlags & TILT_REQUEST_FLAG)))
-		common->wcmFlags |= TILT_REQUEST_FLAG;
-
 	/* pressure curve takes control points x1,y1,x2,y2
 	 * values in range from 0..100.
 	 * Linear curve is 0,0,100,100
@@ -814,18 +924,26 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 		else
 			wcmSetPressureCurve(priv,a,b,c,d);
 	}
+	free(s);
+
+	if (xf86SetBoolOption(pInfo->options, "Pressure2K", 0)) {
+		xf86Msg(X_CONFIG, "%s: Using 2K pressure levels\n", pInfo->name);
+		priv->maxCurve = 2048;
+	}
 
 	/*Serials of tools we want hotpluged*/
 	if (wcmParseSerials (pInfo) != 0)
 		goto error;
 
-	if (IsCursor(priv))
+	if (IsTablet(priv))
 	{
-		common->wcmCursorProxoutDist = xf86SetIntOption(pInfo->options, "CursorProx", 0);
-		if (common->wcmCursorProxoutDist < 0 ||
-				common->wcmCursorProxoutDist > common->wcmMaxDist)
-			xf86Msg(X_CONFIG, "%s: CursorProx invalid %d \n",
-				pInfo->name, common->wcmCursorProxoutDist);
+		const char *prop = IsCursor(priv) ? "CursorProx" : "StylusProx";
+		priv->wcmProxoutDist = xf86SetIntOption(pInfo->options, prop, 0);
+		if (priv->wcmProxoutDist < 0 ||
+				priv->wcmProxoutDist > common->wcmMaxDist)
+			xf86Msg(X_CONFIG, "%s: %s invalid %d \n",
+				pInfo->name, prop, priv->wcmProxoutDist);
+		priv->wcmSurfaceDist = -1;
 	}
 
 	priv->topX = xf86SetIntOption(pInfo->options, "TopX", 0);
@@ -837,6 +955,9 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 	tool = priv->tool;
 	tool->serial = priv->serial;
 
+	common->wcmPanscrollThreshold = xf86SetIntOption(pInfo->options, "PanScrollThreshold",
+			common->wcmPanscrollThreshold);
+
 	/* The first device doesn't need to add any tools/areas as it
 	 * will be the first anyway. So if different, add tool
 	 * and/or area to the existing lists
@@ -907,13 +1028,10 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 	if (TabletHasFeature(common, WCM_2FGT))
 	{
 		int gesture_is_on;
-
-		/* GestureDefault was off for all devices
-		 * except when multi-touch is supported */
-		common->wcmGestureDefault = 1;
+		Bool gesture_default = TabletHasFeature(priv->common, WCM_LCD) ? FALSE : TRUE;
 
 		gesture_is_on = xf86SetBoolOption(pInfo->options, "Gesture",
-					    common->wcmGestureDefault);
+					    gesture_default);
 
 		if (is_primary || IsTouch(priv))
 			common->wcmGesture = gesture_is_on;
@@ -926,6 +1044,12 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 			common->wcmGestureParameters.wcmTapTime);
 	}
 
+	if (IsStylus(priv) || IsEraser(priv)) {
+		common->wcmPressureRecalibration
+			= xf86SetBoolOption(pInfo->options,
+					    "PressureRecalibration", 1);
+	}
+
 	/* Swap stylus buttons 2 and 3 for Tablet PCs */
 	if (TabletHasFeature(common, WCM_TPC) && IsStylus(priv))
 	{
@@ -950,11 +1074,11 @@ error:
 	return FALSE;
 }
 
-/* The values were based on trail and error. */
-#define WCM_BAMBOO3_MAXX 4096.0
-#define WCM_BAMBOO3_ZOOM_DISTANCE 180.0
-#define WCM_BAMBOO3_SCROLL_DISTANCE 80.0
-#define WCM_BAMBOO3_SCROLL_SPREAD_DISTANCE 350.0
+/* The values were based on trial and error with a 3rd-gen Bamboo */
+#define WCM_DEFAULT_MM_XRES           (27.8 * 1000)
+#define WCM_DEFAULT_MM_YRES           (44.5 * 1000)
+#define WCM_ZOOM_DISTANCE_MM          6.5
+#define WCM_SCROLL_DISTANCE_MM        1.8
 
 /**
  * Parse post-init options for this device. Useful for overriding HW
@@ -967,7 +1091,7 @@ error:
  * hotplugging, False if the device is a depent or xorg.conf device.
  * @param is_hotplugged True if the device is a dependent device, FALSE
  * otherwise.
- * @retvalue True on success or False otherwise.
+ * @retval True on success or False otherwise.
  */
 Bool wcmPostInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 			     Bool is_dependent)
@@ -981,10 +1105,10 @@ Bool wcmPostInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 	/* 2FG touch device */
 	if (TabletHasFeature(common, WCM_2FGT) && IsTouch(priv))
 	{
-		int zoom_distance = common->wcmMaxTouchX *
-			(WCM_BAMBOO3_ZOOM_DISTANCE / WCM_BAMBOO3_MAXX);
-		int scroll_distance = common->wcmMaxTouchX *
-			(WCM_BAMBOO3_SCROLL_DISTANCE / WCM_BAMBOO3_MAXX);
+		int x_res = common->wcmTouchResolX ? common->wcmTouchResolX : WCM_DEFAULT_MM_XRES;
+		int y_res = common->wcmTouchResolY ? common->wcmTouchResolY : WCM_DEFAULT_MM_YRES;
+		int zoom_distance = WCM_ZOOM_DISTANCE_MM * x_res / 1000;
+		int scroll_distance = WCM_SCROLL_DISTANCE_MM * y_res / 1000;
 
 		common->wcmGestureParameters.wcmZoomDistance =
 			xf86SetIntOption(pInfo->options, "ZoomDistance",
@@ -993,10 +1117,6 @@ Bool wcmPostInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 		common->wcmGestureParameters.wcmScrollDistance =
 			xf86SetIntOption(pInfo->options, "ScrollDistance",
 					 scroll_distance);
-
-		common->wcmGestureParameters.wcmMaxScrollFingerSpread =
-			common->wcmMaxTouchX *
-			(WCM_BAMBOO3_SCROLL_SPREAD_DISTANCE / WCM_BAMBOO3_MAXX);
 	}
 
 
