@@ -79,29 +79,53 @@ int wcmDevSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode)
 	return wcmDevSwitchModeCall(pInfo, mode);
 }
 
-Atom prop_devnode;
-Atom prop_rotation;
-Atom prop_tablet_area;
-Atom prop_pressurecurve;
-Atom prop_serials;
-Atom prop_serial_binding;
-Atom prop_strip_buttons;
-Atom prop_wheel_buttons;
-Atom prop_tv_resolutions;
-Atom prop_cursorprox;
-Atom prop_threshold;
-Atom prop_suppress;
-Atom prop_touch;
-Atom prop_gesture;
-Atom prop_gesture_param;
-Atom prop_hover;
-Atom prop_tooltype;
-Atom prop_btnactions;
-Atom prop_product_id;
+static Atom prop_devnode;
+static Atom prop_rotation;
+static Atom prop_tablet_area;
+static Atom prop_pressurecurve;
+static Atom prop_serials;
+static Atom prop_serial_binding;
+static Atom prop_strip_buttons;
+static Atom prop_wheel_buttons;
+static Atom prop_proxout;
+static Atom prop_threshold;
+static Atom prop_suppress;
+static Atom prop_touch;
+static Atom prop_hardware_touch;
+static Atom prop_gesture;
+static Atom prop_gesture_param;
+static Atom prop_hover;
+static Atom prop_tooltype;
+static Atom prop_btnactions;
+static Atom prop_product_id;
+static Atom prop_pressure_recal;
+static Atom prop_panscroll_threshold;
 #ifdef DEBUG
-Atom prop_debuglevels;
+static Atom prop_debuglevels;
 #endif
 
+/**
+ * Calculate a user-visible pressure level from a driver-internal pressure
+ * level. Pressure settings exposed to the user assume a range of 0-2047
+ * while the driver scales everything to a range of 0-maxCurve.
+ */
+static inline int wcmInternalToUserPressure(InputInfoPtr pInfo, int pressure)
+{
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	return pressure / (priv->maxCurve / 2048);
+}
+
+/**
+ * Calculate a driver-internal pressure level from a user-visible pressure
+ * level. Pressure settings exposed to the user assume a range of 0-2047
+ * while the driver scales everything to a range of 0-maxCurve.
+ */
+static inline int wcmUserToInternalPressure(InputInfoPtr pInfo, int pressure)
+{
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	return pressure * (priv->maxCurve / 2048);
+}
+
 /**
  * Resets an arbitrary Action property, given a pointer to the old
  * handler and information about the new Action.
@@ -240,8 +264,8 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	}
 
 	values[0] = common->tablet_id;
-	values[1] = priv->old_serial;
-	values[2] = priv->old_device_id;
+	values[1] = priv->oldState.serial_num;
+	values[2] = priv->oldState.device_id;
 	values[3] = priv->cur_serial;
 	values[4] = priv->cur_device_id;
 	prop_serials = InitWcmAtom(pInfo->dev, WACOM_PROP_SERIALIDS, XA_INTEGER, 32, 5, values);
@@ -249,12 +273,13 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	values[0] = priv->serial;
 	prop_serial_binding = InitWcmAtom(pInfo->dev, WACOM_PROP_SERIAL_BIND, XA_INTEGER, 32, 1, values);
 
-	if (IsCursor(priv)) {
-		values[0] = common->wcmCursorProxoutDist;
-		prop_cursorprox = InitWcmAtom(pInfo->dev, WACOM_PROP_PROXIMITY_THRESHOLD, XA_INTEGER, 32, 1, values);
+	if (IsTablet(priv)) {
+		values[0] = priv->wcmProxoutDist;
+		prop_proxout = InitWcmAtom(pInfo->dev, WACOM_PROP_PROXIMITY_THRESHOLD, XA_INTEGER, 32, 1, values);
 	}
 
 	values[0] = (!common->wcmMaxZ) ? 0 : common->wcmThreshold;
+	values[0] = wcmInternalToUserPressure(pInfo, values[0]);
 	prop_threshold = InitWcmAtom(pInfo->dev, WACOM_PROP_PRESSURE_THRESHOLD, XA_INTEGER, 32, 1, values);
 
 	values[0] = common->wcmSuppress;
@@ -264,6 +289,11 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	values[0] = common->wcmTouch;
 	prop_touch = InitWcmAtom(pInfo->dev, WACOM_PROP_TOUCH, XA_INTEGER, 8, 1, values);
 
+	if (common->wcmHasHWTouchSwitch && IsTouch(priv)) {
+		values[0] = common->wcmHWTouchSwitchState;
+		prop_hardware_touch = InitWcmAtom(pInfo->dev, WACOM_PROP_HARDWARE_TOUCH, XA_INTEGER, 8, 1, values);
+	}
+
 	if (IsStylus(priv)) {
 		values[0] = !common->wcmTPCButton;
 		prop_hover = InitWcmAtom(pInfo->dev, WACOM_PROP_HOVER, XA_INTEGER, 8, 1, values);
@@ -300,6 +330,16 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 			wcmResetWheelAction(pInfo, i);
 	}
 
+	if (IsStylus(priv) || IsEraser(priv)) {
+		values[0] = common->wcmPressureRecalibration;
+		prop_pressure_recal = InitWcmAtom(pInfo->dev,
+						  WACOM_PROP_PRESSURE_RECAL,
+						  XA_INTEGER, 8, 1, values);
+	}
+
+	values[0] = common->wcmPanscrollThreshold;
+	prop_panscroll_threshold = InitWcmAtom(pInfo->dev, WACOM_PROP_PANSCROLL_THRESHOLD, XA_INTEGER, 32, 1, values);
+
 	values[0] = common->vendor_id;
 	values[1] = common->tablet_id;
 	prop_product_id = InitWcmAtom(pInfo->dev, XI_PROP_PRODUCT_ID, XA_INTEGER, 32, 2, values);
@@ -320,12 +360,9 @@ static int wcmFindProp(Atom property, Atom *prop_list, int nprops)
 	/* check all properties used for button actions */
 	for (i = 0; i < nprops; i++)
 		if (prop_list[i] == property)
-			break;
+			return i;
 
-	if (i >= nprops)
-		return -BadAtom;
-
-	return i;
+	return -BadAtom;
 }
 
 /**
@@ -337,8 +374,10 @@ static int wcmFindProp(Atom property, Atom *prop_list, int nprops)
  * @param property      The Action property that should be searched for
  * @param[out] handler  Returns a pointer to the property's handler
  * @param[out] action   Returns a pointer to the property's action list
+ * @return              'true' if the property was found. Neither out parameter
+ *                      will be null if this is the case.
  */
-static void wcmFindActionHandler(WacomDevicePtr priv, Atom property, Atom **handler, unsigned int (**action)[256])
+static BOOL wcmFindActionHandler(WacomDevicePtr priv, Atom property, Atom **handler, unsigned int (**action)[256])
 {
 	int offset;
 
@@ -347,7 +386,7 @@ static void wcmFindActionHandler(WacomDevicePtr priv, Atom property, Atom **hand
 	{
 		*handler = &priv->btn_actions[offset];
 		*action  = &priv->keys[offset];
-		return;
+		return TRUE;
 	}
 
 	offset = wcmFindProp(property, priv->wheel_actions, ARRAY_SIZE(priv->wheel_actions));
@@ -355,28 +394,49 @@ static void wcmFindActionHandler(WacomDevicePtr priv, Atom property, Atom **hand
 	{
 		*handler = &priv->wheel_actions[offset];
 		*action  = &priv->wheel_keys[offset];
-		return;
+		return TRUE;
 	}
 
 	offset = wcmFindProp(property, priv->strip_actions, ARRAY_SIZE(priv->strip_actions));
-	if (offset < 0)
+	if (offset >= 0)
 	{
 		*handler = &priv->strip_actions[offset];
 		*action  = &priv->strip_keys[offset];
-		return;
+		return TRUE;
 	}
+
+	return FALSE;
 }
 
-static int wcmCheckActionProperty(Atom property, XIPropertyValuePtr prop)
+static int wcmCheckActionProperty(WacomDevicePtr priv, Atom property, XIPropertyValuePtr prop)
 {
 	CARD32 *data;
 	int j;
 
-	if (!property)
-		return Success;
-
-	if (prop->size >= 255 || prop->format != 32 || prop->type != XA_INTEGER)
+	if (!property) {
+		DBG(3, priv, "ERROR: Atom is NONE\n");
 		return BadMatch;
+	}
+
+	if (prop == NULL) {
+		DBG(3, priv, "ERROR: Value is NULL\n");
+		return BadMatch;
+	}
+
+	if (prop->size >= 255) {
+		DBG(3, priv, "ERROR: Too many values (%ld > 255)\n", prop->size);
+		return BadMatch;
+	}
+
+	if (prop->format != 32) {
+		DBG(3, priv, "ERROR: Incorrect value format (%d != 32)\n", prop->format);
+		return BadMatch;
+	}
+
+	if (prop->type != XA_INTEGER) {
+		DBG(3, priv, "ERROR: Incorrect value type (%d != XA_INTEGER)\n", prop->type);
+		return BadMatch;
+	}
 
 	data = (CARD32*)prop->data;
 
@@ -385,18 +445,24 @@ static int wcmCheckActionProperty(Atom property, XIPropertyValuePtr prop)
 		int code = data[j] & AC_CODE;
 		int type = data[j] & AC_TYPE;
 
+		DBG(10, priv, "Index %d == %d (type: %d, code: %d)\n", j, data[j], type, code);
+
 		switch(type)
 		{
 			case AC_KEY:
 				break;
 			case AC_BUTTON:
-				if (code > WCM_MAX_BUTTONS)
+				if (code > WCM_MAX_X11BUTTON) {
+					DBG(3, priv, "ERROR: AC_BUTTON code too high (%d > %d)\n", code, WCM_MAX_X11BUTTON);
 					return BadValue;
+				}
 				break;
-			case AC_DISPLAYTOGGLE:
 			case AC_MODETOGGLE:
 				break;
+			case AC_PANSCROLL:
+				break;
 			default:
+				DBG(3, priv, "ERROR: Unknown command\n");
 				return BadValue;
 		}
 	}
@@ -425,13 +491,21 @@ static int wcmSetActionProperty(DeviceIntPtr dev, Atom property,
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	int rc, i;
 
-	DBG(10, priv, "\n");
+	DBG(5, priv, "%s new actions for Atom %d\n", checkonly ? "Checking" : "Setting", property);
 
-	rc = wcmCheckActionProperty(property, prop);
-	if (rc != Success)
+	rc = wcmCheckActionProperty(priv, property, prop);
+	if (rc != Success) {
+		const char *msg = NULL;
+		switch (rc) {
+			case BadMatch: msg = "BadMatch"; break;
+			case BadValue: msg = "BadValue"; break;
+			default: msg = "UNKNOWN"; break;
+		}
+		DBG(3, priv, "Action validation failed with code %d (%s)\n", rc, msg);
 		return rc;
+	}
 
-	if (!checkonly && prop)
+	if (!checkonly)
 	{
 		memset(action, 0, sizeof(*action));
 		for (i = 0; i < prop->size; i++)
@@ -485,7 +559,10 @@ static int wcmSetActionsProperty(DeviceIntPtr dev, Atom property,
                                  int size, Atom* handlers, unsigned int (*actions)[256])
 {
 	InputInfoPtr pInfo = (InputInfoPtr) dev->public.devicePrivate;
-	int rc;
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
+	int rc, i;
+
+	DBG(10, priv, "\n");
 
 	if (prop->size != size)
 		return BadValue;
@@ -494,28 +571,26 @@ static int wcmSetActionsProperty(DeviceIntPtr dev, Atom property,
 	if (rc != Success)
 		return rc;
 
-	if (!checkonly)
+	for (i = 0; i < prop->size; i++)
 	{
-		int i;
+		int index = i;
+		Atom subproperty = ((Atom*)prop->data)[i];
+		XIPropertyValuePtr subprop;
 
-		for (i = 0; i < prop->size; i++)
-		{
-			int index = i;
-			Atom subproperty = ((Atom*)prop->data)[i];
-			XIPropertyValuePtr subprop;
+		if (property == prop_btnactions)
+		{ /* Driver uses physical -- not X11 -- button numbering internally */
+			if (i < 3)
+				index = i;
+			else if (i < 7)
+				continue;
+			else
+				index = i - 4;
+		}
 
-			if (property == prop_btnactions)
-			{ /* Driver uses physical -- not X11 -- button numbering internally */
-				if (i < 3)
-					index = i;
-				else if (i < 7)
-					continue;
-				else
-					index = i - 4;
-			}
-
-			if (subproperty == 0)
-			{ /* Interpret 'None' as meaning 'reset' */
+		if (subproperty == 0)
+		{ /* Interpret 'None' as meaning 'reset' */
+			if (!checkonly)
+			{
 				if (property == prop_btnactions)
 					wcmResetButtonAction(pInfo, index, size);
 				else if (property == prop_strip_buttons)
@@ -526,9 +601,13 @@ static int wcmSetActionsProperty(DeviceIntPtr dev, Atom property,
 				if (subproperty != handlers[index])
 					subproperty = handlers[index];
 			}
-
+		}
+		else
+		{
 			XIGetDeviceProperty(dev, subproperty, &subprop);
-			wcmSetActionProperty(dev, subproperty, subprop, checkonly, &handlers[index], &actions[index]);
+			rc = wcmSetActionProperty(dev, subproperty, subprop, checkonly, &handlers[index], &actions[index]);
+			if (rc != Success)
+				return rc;
 		}
 	}
 
@@ -562,6 +641,65 @@ void wcmUpdateRotationProperty(WacomDevicePtr priv)
 	}
 }
 
+static void
+wcmSetHWTouchProperty(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = pInfo->private;
+	WacomCommonPtr common = priv->common;
+	XIPropertyValuePtr prop;
+	CARD8 prop_value;
+	int rc;
+
+	rc = XIGetDeviceProperty(pInfo->dev, prop_hardware_touch, &prop);
+	if (rc != Success || prop->format != 8 || prop->size != 1)
+	{
+		xf86Msg(X_ERROR, "%s: Failed to update hardware touch state.\n",
+			pInfo->name);
+		return;
+	}
+
+	prop_value = common->wcmHWTouchSwitchState;
+	XIChangeDeviceProperty(pInfo->dev, prop_hardware_touch, XA_INTEGER,
+			       prop->format, PropModeReplace,
+			       prop->size, &prop_value, TRUE);
+}
+
+static CARD32
+touchTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
+{
+	InputInfoPtr pInfo = arg;
+#if !HAVE_THREADED_INPUT
+	int sigstate = xf86BlockSIGIO();
+#endif
+
+	wcmSetHWTouchProperty(pInfo);
+
+#if !HAVE_THREADED_INPUT
+	xf86UnblockSIGIO(sigstate);
+#endif
+
+	return 0;
+}
+
+/**
+ * Update HW touch property when its state is changed by touch switch
+ */
+void
+wcmUpdateHWTouchProperty(WacomDevicePtr priv, int hw_touch)
+{
+	WacomCommonPtr common = priv->common;
+
+	if (hw_touch == common->wcmHWTouchSwitchState)
+		return;
+
+	common->wcmHWTouchSwitchState = hw_touch;
+
+	/* This function is called during SIGIO/InputThread. Schedule timer
+	 * for property event delivery by the main thread. */
+	priv->touch_timer = TimerSet(priv->touch_timer, 0 /* reltime */,
+				      1, touchTimerFunc, priv->pInfo);
+}
+
 /**
  * Only allow deletion of a property if it is not being used by any of the
  * button actions.
@@ -606,8 +744,8 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 			if ((values[0] == -1) && (values[1] == -1) &&
 					(values[2] == -1) && (values[3] == -1))
 			{
-				values[0] = 0;
-				values[1] = 0;
+				values[0] = priv->minX;
+				values[1] = priv->minX;
 				values[2] = priv->maxX;
 				values[3] = priv->maxY;
 			}
@@ -697,14 +835,14 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 		return wcmSetActionsProperty(dev, property, prop, checkonly, ARRAY_SIZE(priv->strip_actions), priv->strip_actions, priv->strip_keys);
 	else if (property == prop_wheel_buttons)
 		return wcmSetActionsProperty(dev, property, prop, checkonly, ARRAY_SIZE(priv->wheel_actions), priv->wheel_actions, priv->wheel_keys);
-	else if (property == prop_cursorprox)
+	else if (property == prop_proxout)
 	{
 		CARD32 value;
 
 		if (prop->size != 1 || prop->format != 32)
 			return BadValue;
 
-		if (!IsCursor (priv))
+		if (!IsTablet (priv))
 			return BadValue;
 
 		value = *(CARD32*)prop->data;
@@ -713,9 +851,10 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 			return BadValue;
 
 		if (!checkonly)
-			common->wcmCursorProxoutDist = value;
+			priv->wcmProxoutDist = value;
 	} else if (property == prop_threshold)
 	{
+		const INT32 MAXIMUM = wcmInternalToUserPressure(pInfo, priv->maxCurve);
 		INT32 value;
 
 		if (prop->size != 1 || prop->format != 32)
@@ -724,9 +863,11 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 		value = *(INT32*)prop->data;
 
 		if (value == -1)
-			value = DEFAULT_THRESHOLD;
-		else if ((value < 1) || (value > FILTER_PRESSURE_RES))
+			value = priv->maxCurve * DEFAULT_THRESHOLD;
+		else if ((value < 1) || (value > MAXIMUM))
 			return BadValue;
+		else
+			value = wcmUserToInternalPressure(pInfo, value);
 
 		if (!checkonly)
 			common->wcmThreshold = value;
@@ -742,6 +883,19 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 
 		if (!checkonly && common->wcmTouch != values[0])
 			common->wcmTouch = values[0];
+	} else if (property == prop_hardware_touch)
+	{
+		if (common->wcmHasHWTouchSwitch)
+		{
+			/* If we get here from wcmUpdateHWTouchProperty, we know
+			 * the wcmHWTouchSwitchState has been set internally
+			 * already, so we can reply with success. */
+			if (prop->size == 1 && prop->format == 8)
+				if (((CARD8*)prop->data)[0] == common->wcmHWTouchSwitchState)
+					return Success;
+		}
+
+		return BadValue; /* read-only */
 	} else if (property == prop_gesture)
 	{
 		CARD8 *values = (CARD8*)prop->data;
@@ -809,13 +963,42 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 	{
 		int nbuttons = priv->nbuttons < 4 ? priv->nbuttons : priv->nbuttons + 4;
 		return wcmSetActionsProperty(dev, property, prop, checkonly, nbuttons, priv->btn_actions, priv->keys);
+	} else if (property == prop_pressure_recal)
+	{
+		CARD8 *values = (CARD8*)prop->data;
+
+		if (prop->size != 1 || prop->format != 8)
+			return BadValue;
+
+		if ((values[0] != 0) && (values[0] != 1))
+			return BadValue;
+
+		if (!IsStylus(priv) && !IsEraser(priv))
+			return BadMatch;
+
+		if (!checkonly)
+			common->wcmPressureRecalibration = values[0];
+	} else if (property == prop_panscroll_threshold)
+	{
+		CARD32 *values = (CARD32*)prop->data;
+
+		if (prop->size != 1 || prop->format != 32)
+			return BadValue;
+
+		if (values[0] <= 0)
+			return BadValue;
+
+		if (IsTouch(priv))
+			return BadMatch;
+
+		if (!checkonly)
+			common->wcmPanscrollThreshold = values[0];
 	} else
 	{
 		Atom *handler = NULL;
 		unsigned int (*action)[256] = NULL;
-		wcmFindActionHandler(priv, property, &handler, &action);
-		if (handler != NULL && action != NULL)
-			wcmSetActionProperty(dev, property, prop, checkonly, handler, action);
+		if (wcmFindActionHandler(priv, property, &handler, &action))
+			return wcmSetActionProperty(dev, property, prop, checkonly, handler, action);
 		/* backwards-compatible behavior silently ignores the not-found case */
 	}
 
@@ -835,12 +1018,12 @@ int wcmGetProperty (DeviceIntPtr dev, Atom property)
 		uint32_t values[5];
 
 		values[0] = common->tablet_id;
-		values[1] = priv->old_serial;
-		values[2] = priv->old_device_id;
+		values[1] = priv->oldState.serial_num;
+		values[2] = priv->oldState.device_id;
 		values[3] = priv->cur_serial;
 		values[4] = priv->cur_device_id;
 
-		DBG(10, priv, "Update to serial: %d\n", priv->old_serial);
+		DBG(10, priv, "Update to serial: %d\n", priv->oldState.serial_num);
 
 		return XIChangeDeviceProperty(dev, property, XA_INTEGER, 32,
 					      PropModeReplace, 5,
@@ -887,24 +1070,20 @@ int wcmGetProperty (DeviceIntPtr dev, Atom property)
 	return Success;
 }
 
-static CARD32
-serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
+static void
+wcmSetSerialProperty(InputInfoPtr pInfo)
 {
-	InputInfoPtr pInfo = arg;
 	WacomDevicePtr priv = pInfo->private;
 	XIPropertyValuePtr prop;
 	CARD32 prop_value[5];
-	int sigstate;
 	int rc;
 
-	sigstate = xf86BlockSIGIO();
-
 	rc = XIGetDeviceProperty(pInfo->dev, prop_serials, &prop);
 	if (rc != Success || prop->format != 32 || prop->size != 5)
 	{
 		xf86Msg(X_ERROR, "%s: Failed to update serial number.\n",
 			pInfo->name);
-		return 0;
+		return;
 	}
 
 	memcpy(prop_value, prop->data, sizeof(prop_value));
@@ -914,8 +1093,22 @@ serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
 	XIChangeDeviceProperty(pInfo->dev, prop_serials, XA_INTEGER,
 			       prop->format, PropModeReplace,
 			       prop->size, prop_value, TRUE);
+}
 
+static CARD32
+serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
+{
+	InputInfoPtr pInfo = arg;
+
+#if !HAVE_THREADED_INPUT
+	int sigstate = xf86BlockSIGIO();
+#endif
+
+	wcmSetSerialProperty(pInfo);
+
+#if !HAVE_THREADED_INPUT
 	xf86UnblockSIGIO(sigstate);
+#endif
 
 	return 0;
 }
@@ -931,8 +1124,8 @@ wcmUpdateSerial(InputInfoPtr pInfo, unsigned int serial, int id)
 	priv->cur_serial = serial;
 	priv->cur_device_id = id;
 
-	/* This function is called during SIGIO. Schedule timer for property
-	 * event delivery outside of signal handler. */
+	/* This function is called during SIGIO/InputThread. Schedule timer
+	 * for property event delivery by the main thread. */
 	priv->serial_timer = TimerSet(priv->serial_timer, 0 /* reltime */,
 				      1, serialTimerFunc, pInfo);
 }
