@@ -1,6 +1,6 @@
 /*
  * Copyright 1995-2002 by Frederic Lepied, France. <Lepied@XFree86.org>
- * Copyright 2002-2010 by Ping Cheng, Wacom. <pingc@wacom.com>
+ * Copyright 2002-2013 by Ping Cheng, Wacom. <pingc@wacom.com>
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -26,7 +26,6 @@
 #include <wacom-util.h>
 #include <asm/types.h>
 #include <linux/input.h>
-#define MAX_USB_EVENTS 32
 
 /* vendor IDs on the kernel device */
 #define WACOM_VENDOR_ID 0x056a
@@ -48,9 +47,6 @@
 #define TILT_MIN -64		/* Minimum reported tilt value */
 #define TILT_MAX 63		/* Maximum reported tilt value */
 
-#define MIN_PAD_RING 0		/* I4 absolute scroll ring min value */
-#define MAX_PAD_RING 71		/* I4 absolute scroll ring max value */
-
 /* I4 cursor tool has a rotation offset of 175 degrees */
 #define INTUOS4_CURSOR_ROTATION_OFFSET 175
 
@@ -96,9 +92,11 @@
 #define ABS_MT_SLOT 0x2f
 #endif
 
-/* defines to discriminate second side button and the eraser */
-#define ERASER_PROX     4
-#define OTHER_PROX      1
+/* 4.15 */
+
+#ifndef BTN_STYLUS3
+#define BTN_STYLUS3 0x149
+#endif
 
 /******************************************************************************
  * Forward Declarations
@@ -112,7 +110,6 @@ typedef struct _WacomCommonRec WacomCommonRec, *WacomCommonPtr;
 typedef struct _WacomFilterState WacomFilterState, *WacomFilterStatePtr;
 typedef struct _WacomDeviceClass WacomDeviceClass, *WacomDeviceClassPtr;
 typedef struct _WacomTool WacomTool, *WacomToolPtr;
-typedef struct _WacomToolArea WacomToolArea, *WacomToolAreaPtr;
 
 /******************************************************************************
  * WacomModel - model-specific device capabilities
@@ -171,12 +168,14 @@ struct _WacomModel
 							  always an LCD) */
 #define WCM_PENTOUCH		0x00000400 /* Tablet supports pen and touch */
 #define WCM_DUALRING		0x00000800 /* Tablet has two touch rings */
+#define WCM_LEGACY_IDS		0x00001000 /* Tablet uses legacy device IDs */
 #define TabletHasFeature(common, feature) MaskIsSet((common)->tablet_type, (feature))
 #define TabletSetFeature(common, feature) MaskSet((common)->tablet_type, (feature))
 
 #define ABSOLUTE_FLAG		0x00000100
 #define BAUD_19200_FLAG		0x00000400
 #define BUTTONS_ONLY_FLAG	0x00000800
+#define SCROLLMODE_FLAG		0x00001000
 
 #define IsCursor(priv) (DEVICE_ID((priv)->flags) == CURSOR_ID)
 #define IsStylus(priv) (DEVICE_ID((priv)->flags) == STYLUS_ID)
@@ -184,15 +183,17 @@ struct _WacomModel
 #define IsEraser(priv) (DEVICE_ID((priv)->flags) == ERASER_ID)
 #define IsPad(priv)    (DEVICE_ID((priv)->flags) == PAD_ID)
 #define IsPen(priv)    (IsStylus(priv) || IsEraser(priv))
+#define IsTablet(priv) (IsPen(priv) || IsCursor(priv))
 
 #define IsUSBDevice(common) ((common)->wcmDevCls == &gWacomUSBDevice)
 
-#define FILTER_PRESSURE_RES	2048	/* maximum points in pressure curve */
+#define FILTER_PRESSURE_RES	65536	/* maximum points in pressure curve */
 /* Tested result for setting the pressure threshold to a reasonable value */
-#define THRESHOLD_TOLERANCE (FILTER_PRESSURE_RES / 125)
-#define DEFAULT_THRESHOLD (FILTER_PRESSURE_RES / 75)
+#define THRESHOLD_TOLERANCE (0.008f)
+#define DEFAULT_THRESHOLD (0.013f)
 
 #define WCM_MAX_BUTTONS		32	/* maximum number of tablet buttons */
+#define WCM_MAX_X11BUTTON	127	/* maximum button number X11 can handle */
 
 #define AXIS_INVERT  0x01               /* Flag describing an axis which increases "downward" */
 #define AXIS_BITWISE 0x02               /* Flag describing an axis which changes bitwise */
@@ -209,17 +210,43 @@ struct _WacomModel
 #define STRIP_RIGHT_UP    2
 #define STRIP_RIGHT_DN    3
 
-/* get/set/property */
-typedef struct _PROPINFO PROPINFO;
+#define IDX_KEY_CONTROLPANEL		1
+#define IDX_KEY_ONSCREEN_KEYBOARD	2
+#define IDX_KEY_BUTTONCONFIG		3
+#define IDX_KEY_INFO			4
 
-struct _PROPINFO
+/******************************************************************************
+ * WacomDeviceState
+ *****************************************************************************/
+struct _WacomDeviceState
 {
-	Atom wcmProp;
-	char paramName[32];
-	int nParamID;
-	int nFormat;
-	int nSize;
-	int nDefault;
+	InputInfoPtr pInfo;
+	int device_id;		/* tool id reported from the physical device */
+	int device_type;
+	unsigned int serial_num;
+	int x;
+	int y;
+	int buttons;
+	int pressure;
+	int tiltx;
+	int tilty;
+	int stripx;
+	int stripy;
+	int rotation;
+	int abswheel;
+	int abswheel2;
+	int relwheel;
+	int distance;
+	int throttle;
+	int proximity;
+	int sample;	/* wraps every 24 days */
+	int time;
+	unsigned int keys; /* bitmask for IDX_KEY_CONTROLPANEL, etc. */
+};
+
+static const struct _WacomDeviceState OUTPROX_STATE = {
+  .abswheel = INT_MAX,
+  .abswheel2 = INT_MAX
 };
 
 struct _WacomDeviceRec
@@ -237,17 +264,13 @@ struct _WacomDeviceRec
 	int bottomY;		/* Y bottom in device coordinates */
 	int resolX;             /* X resolution */
 	int resolY;             /* Y resolution */
-	int maxX;	        /* tool physical maxX in device coordinates*/
-	int maxY;	        /* tool physical maxY in device coordinates*/
-	double factorX;		/* X factor */
-	double factorY;		/* Y factor */
+	int minX;	        /* tool physical minX in device coordinates */
+	int minY;	        /* tool physical minY in device coordinates */
+	int maxX;	        /* tool physical maxX in device coordinates */
+	int maxY;	        /* tool physical maxY in device coordinates */
 	unsigned int serial;	/* device serial number this device takes (if 0, any serial is ok) */
 	unsigned int cur_serial; /* current serial in prox */
 	int cur_device_id;	/* current device ID in prox */
-	int maxWidth;		/* max active screen width in screen coords */
-	int maxHeight;		/* max active screen height in screen coords */
-	int leftPadding;	/* left padding for virtual tablet in device coordinates*/
-	int topPadding;		/* top padding for virtual tablet in device coordinates*/
 
 	/* button mapping information
 	 *
@@ -272,77 +295,31 @@ struct _WacomDeviceRec
 	WacomCommonPtr common;  /* common info pointer */
 
 	/* state fields in device coordinates */
-	int currentX;           /* current X position */
-	int currentY;           /* current Y position */
-	int currentSX;          /* current screen X position in screen coords */
-	int currentSY;          /* current screen Y position in screen coords */
-	int oldX;               /* previous X position */
-	int oldY;               /* previous Y position */
-	int oldZ;               /* previous pressure */
-	int oldTiltX;           /* previous tilt in x direction */
-	int oldTiltY;           /* previous tilt in y direction */    
-	int oldWheel;           /* previous wheel value */    
-	int oldWheel2;          /* previous wheel2 value */
-	int oldRot;             /* previous rotation value */
-	int oldStripX;          /* previous left strip value */
-	int oldStripY;          /* previous right strip value */
-	int oldThrottle;        /* previous throttle value */
-	int oldButtons;         /* previous buttons state */
-	int oldProximity;       /* previous proximity */
+	struct _WacomDeviceState wcmPanscrollState; /* panscroll state tracking */
+	struct _WacomDeviceState oldState; /* previous state information */
 	int oldCursorHwProx;	/* previous cursor hardware proximity */
-	int old_device_id;	/* last in prox device id */
-	unsigned int old_serial;/* last in prox tool serial number */
-	int devReverseCount;	/* Relative ReverseConvert called twice each movement*/
 
-	/* JEJ - throttle */
-	int throttleStart;      /* time in ticks for last wheel movement */
-	int throttleLimit;      /* time in ticks for next wheel movement */
-	int throttleValue;      /* current throttle value */
-
-	/* JEJ - filters */
-	int pPressCurve[FILTER_PRESSURE_RES + 1]; /* pressure curve */
+	int maxCurve;		/* maximum pressure curve value */
+	int *pPressCurve;       /* pressure curve */
 	int nPressCtrl[4];      /* control points for curve */
 	int minPressure;	/* the minimum pressure a pen may hold */
-
+	int oldMinPressure;     /* to record the last minPressure before going out of proximity */
+	int wcmSurfaceDist;	/* Distance reported by hardware when tool at surface */
+	int wcmProxoutDist;     /* Distance from surface when proximity-out should be triggered */
+	unsigned int eventCnt;  /* count number of events while in proximity */
+	int maxRawPressure;     /* maximum 'raw' pressure seen until first button event */
 	WacomToolPtr tool;         /* The common tool-structure for this device */
 
 	int isParent;		/* set to 1 if the device is not auto-hotplugged */
 
 	OsTimerPtr serial_timer; /* timer used for serial number property update */
 	OsTimerPtr tap_timer;   /* timer used for tap timing */
+	OsTimerPtr touch_timer; /* timer used for touch switch property update */
 };
 
-/******************************************************************************
- * WacomDeviceState
- *****************************************************************************/
-
 #define MAX_SAMPLES	20
 #define DEFAULT_SAMPLES 4
 
-struct _WacomDeviceState
-{
-	InputInfoPtr pInfo;
-	int device_id;		/* tool id reported from the physical device */
-	int device_type;
-	unsigned int serial_num;
-	int x;
-	int y;
-	int buttons;
-	int pressure;
-	int tiltx;
-	int tilty;
-	int stripx;
-	int stripy;
-	int rotation;
-	int abswheel;
-	int abswheel2;
-	int relwheel;
-	int distance;
-	int throttle;
-	int proximity;
-	int sample;	/* wraps every 24 days */
-};
-
 struct _WacomFilterState
 {
         int npoints;
@@ -383,7 +360,7 @@ struct _WacomDeviceClass
 {
 	Bool (*Detect)(InputInfoPtr pInfo); /* detect device */
 	Bool (*ParseOptions)(InputInfoPtr pInfo); /* parse class-specific options */
-	Bool (*Init)(InputInfoPtr pInfo, char* id, float *version);   /* initialize device */
+	Bool (*Init)(InputInfoPtr pInfo, char* id, size_t id_len, float *version);   /* initialize device */
 	int  (*ProbeKeys)(InputInfoPtr pInfo); /* set the bits for the keys supported */
 };
 
@@ -394,18 +371,16 @@ extern WacomDeviceClass gWacomISDV4Device;
  * WacomCommonRec
  *****************************************************************************/
 
-#define TILT_REQUEST_FLAG       1
 #define TILT_ENABLED_FLAG       2
 
-#define MAX_CHANNELS 17
+#define MAX_FINGERS 16
+#define MAX_CHANNELS (MAX_FINGERS+2) /* one channel for stylus/mouse. The other one for pad */
 #define PAD_CHANNEL (MAX_CHANNELS-1)
-#define MAX_FINGERS (MAX_CHANNELS-1)
 
 typedef struct {
 	int wcmZoomDistance;	       /* minimum distance for a zoom touch gesture */
 	int wcmScrollDistance;	       /* minimum motion before sending a scroll gesture */
 	int wcmScrollDirection;	       /* store the vertical or horizontal bit in use */
-	int wcmMaxScrollFingerSpread; /* maximum distance between fingers for scroll gesture */
 	int wcmGestureUsed;	       /* retain used gesture count within one in-prox event */
 	int wcmTapTime;	   	       /* minimum time between taps for a right click */
 } WacomGesturesParameters;
@@ -416,10 +391,16 @@ enum WacomProtocol {
 	WCM_PROTOCOL_5
 };
 
+struct _WacomDriverRec
+{
+	WacomDevicePtr active;     /* Arbitrate motion through this pointer */
+};
+extern struct _WacomDriverRec WACOM_DRIVER; // Defined in wcmCommon.c
+
 struct _WacomCommonRec 
 {
 	/* Do not move device_path, same offset as priv->name. Used by DBG macro */
-	const char* device_path;    /* device file name */
+	char* device_path;          /* device file name */
 	dev_t min_maj;               /* minor/major number */
 	unsigned char wcmFlags;     /* various flags (handle tilt) */
 	int debugLevel;
@@ -431,9 +412,13 @@ struct _WacomCommonRec
 	unsigned long wcmKeys[NBITS(KEY_MAX)]; /* supported tool types for the device */
 	WacomDevicePtr wcmTouchDevice; /* The pointer for pen to access the
 					  touch tool of the same device id */
-	Bool wcmPenInProx;      /* Keep pen in-prox state for touch tool */
+
+	Bool wcmHasHWTouchSwitch;    /* Tablet has a touch on/off switch */
+	int wcmHWTouchSwitchState;   /* touch event disable/enabled by hardware switch */
 
 	/* These values are in tablet coordinates */
+	int wcmMinX;                 /* tablet min X value */
+	int wcmMinY;                 /* tablet min Y value */
 	int wcmMaxX;                 /* tablet max X value */
 	int wcmMaxY;                 /* tablet max Y value */
 	int wcmMaxZ;                 /* tablet max Z value */
@@ -463,6 +448,8 @@ struct _WacomCommonRec
 
 	int wcmMaxStripX;            /* Maximum fingerstrip X */
 	int wcmMaxStripY;            /* Maximum fingerstrip Y */
+	int wcmMinRing;              /* Minimum touchring value */
+	int wcmMaxRing;              /* Maximum touchring value */
 
 	WacomDevicePtr wcmDevices;   /* list of devices sharing same port */
 	int wcmPktLength;            /* length of a packet */
@@ -478,15 +465,15 @@ struct _WacomCommonRec
 	int wcmTouch;	             /* disable/enable touch event */
 	int wcmTouchDefault;	     /* default to disable when not supported */
 	int wcmGesture;	     	     /* disable/enable touch gesture */
-	int wcmGestureDefault;       /* default touch gesture to disable when not supported */
 	int wcmGestureMode;	       /* data is in Gesture Mode? */
 	WacomDeviceState wcmGestureState[MAX_FINGERS]; /* inital state when in gesture mode */
 	WacomGesturesParameters wcmGestureParameters;
-	int wcmMaxCursorDist;	     /* Max mouse distance reported so far */
-	int wcmCursorProxoutDist;    /* Max mouse distance for proxy-out max/256 units */
-	int wcmCursorProxoutDistDefault; /* Default max mouse distance for proxy-out */
+	int wcmProxoutDistDefault;   /* Default value for wcmProxoutDist */
 	int wcmSuppress;        	 /* transmit position on delta > supress */
 	int wcmRawSample;	     /* Number of raw data used to filter an event */
+	int wcmPressureRecalibration; /* Determine if pressure recalibration of
+					 worn pens should be performed */
+	int wcmPanscrollThreshold;	/* distance pen must move to send a panscroll event */
 
 	int bufpos;                        /* position with buffer */
 	unsigned char buffer[BUFFER_SIZE]; /* data read from device */
