@@ -82,31 +82,26 @@ static void
 test_get_wheel_button(void)
 {
 	int delta;
-	int button_up, button_dn, action_up, action_dn;
+	int action_up, action_dn;
 
-	button_up = 100;
-	button_dn = 200;
 	action_up = 300;
 	action_dn = 400;
 
 	for (delta = -32; delta <= 32; delta++)
 	{
-		int *action;
-		int result = getWheelButton(delta, button_up, button_dn, &action_up, &action_dn, &action);
+		int action;
+		action = getWheelButton(delta, action_up, action_dn);
 		if (delta < 0)
 		{
-			assert(result == button_dn);
-			assert(action == &action_dn);
+			assert(action == action_dn);
 		}
 		else if (delta == 0)
 		{
-			assert(result == 0);
-			assert(action == NULL);
+			assert(action == -1);
 		}
 		else
 		{
-			assert(result == button_up);
-			assert(action == &action_up);
+			assert(action == action_up);
 		}
 	}
 }
@@ -157,7 +152,7 @@ test_rebase_pressure(void)
 	ds.pressure = 10;
 
 	/* Pressure in out-of-proximity means get new preloaded pressure */
-	priv.oldProximity = 0;
+	priv.oldState.proximity = 0;
 
 	/* make sure we don't touch priv, not really needed, the compiler should
 	 * honor the consts but... */
@@ -169,7 +164,7 @@ test_rebase_pressure(void)
 	assert(memcmp(&priv, &base, sizeof(priv)) == 0);
 
 	/* Pressure in-proximity means rebase to new minimum */
-	priv.oldProximity = 1;
+	priv.oldState.proximity = 1;
 
 	base = priv;
 
@@ -184,54 +179,59 @@ test_normalize_pressure(void)
 	InputInfoRec pInfo = {0};
 	WacomDeviceRec priv = {0};
 	WacomCommonRec common = {0};
-	WacomDeviceState ds = {0};
+	int normalized_max = 65536;
 	int pressure, prev_pressure = -1;
-	int i, j;
+	int i, j, k;
 
 	priv.common = &common;
 	priv.pInfo = &pInfo;
-	pInfo.name = "Wacom test device";
+	pInfo.name = strdupa("Wacom test device");
+	common.wcmPressureRecalibration = 1;
 
 	priv.minPressure = 0;
 
-	/* Some random loop to check various maxZ pressure values. Starting at
-	 * 1, because if wcmMaxZ is 0 we have other problems. */
-	for (j = 1; j <= 256; j += 17)
-	{
-		common.wcmMaxZ = j;
-		prev_pressure = -1;
+	/* Check various maxCurve values */
+	for (k = 512; k <= normalized_max; k += 239) {
+		priv.maxCurve = k;
 
-		for (i = 0; i <= common.wcmMaxZ; i++)
+		/* Some random loop to check various maxZ pressure values. Starting at
+		 * 1, because if wcmMaxZ is 0 we have other problems. */
+		for (j = 1; j <= 256; j += 17)
 		{
-			ds.pressure = i;
+			common.wcmMaxZ = j;
+			prev_pressure = -1;
 
-			pressure = normalizePressure(&priv, &ds);
-			assert(pressure >= 0);
-			assert(pressure <= FILTER_PRESSURE_RES);
+			for (i = 0; i <= common.wcmMaxZ; i++)
+			{
+				pressure = i;
 
-			/* we count up, so assume normalised pressure goes up too */
-			assert(prev_pressure < pressure);
-			prev_pressure = pressure;
+				pressure = normalizePressure(&priv, pressure);
+				assert(pressure >= 0);
+				assert(pressure <= k);
+
+				/* we count up, so assume normalised pressure goes up too */
+				assert(prev_pressure < pressure);
+				prev_pressure = pressure;
+			}
+
+			assert(pressure == k);
 		}
-
-		assert(pressure == FILTER_PRESSURE_RES);
 	}
 
 	/* If minPressure is higher than ds->pressure, normalizePressure takes
 	 * minPressure and ignores actual pressure. This would be a bug in the
 	 * driver code, but we might as well test for it. */
 	priv.minPressure = 10;
-	ds.pressure = 0;
+	priv.maxCurve = normalized_max;
 
-	prev_pressure = normalizePressure(&priv, &ds);
+	prev_pressure = normalizePressure(&priv, 0);
 	for (i = 0; i < priv.minPressure; i++)
 	{
-		ds.pressure = i;
 
-		pressure = normalizePressure(&priv, &ds);
+		pressure = normalizePressure(&priv, i);
 
 		assert(pressure >= 0);
-		assert(pressure < FILTER_PRESSURE_RES);
+		assert(pressure < normalized_max);
 
 		/* we count up, so assume normalised pressure goes up too */
 		assert(prev_pressure == pressure);
@@ -252,7 +252,9 @@ test_initial_size(void)
 	WacomDeviceRec priv = {0};
 	WacomCommonRec common = {0};
 
-	int minx, maxx, miny, maxy, xres, yres;
+	/* pin to some numbers */
+	int xres = 1920, yres = 1600;
+	int minx, maxx = 2 * xres, miny, maxy = 2 * yres;
 
 	info.private = &priv;
 	priv.common = &common;
