@@ -28,6 +28,7 @@
 #include <limits.h>
 #include <stdio.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <ctype.h>
 #include <stdlib.h>
 #include <getopt.h>
@@ -471,6 +472,15 @@ static param_t parameters[] =
 		.arg_count = 1,
 		.prop_flags = PROP_FLAG_BOOLEAN
 	},
+	{
+		.name = "PanScrollThreshold",
+		.x11name = "PanScrollThreshold",
+		.desc = "Adjusts distance required for pan actions to generate a scroll event",
+		.prop_name = WACOM_PROP_PANSCROLL_THRESHOLD,
+		.prop_format = 32,
+		.prop_offset = 0,
+		.arg_count = 1,
+	},
 	{
 		.name = "MapToOutput",
 		.desc = "Map the device to the given output. ",
@@ -484,7 +494,7 @@ static param_t parameters[] =
 		.get_func = get_all,
 		.prop_flags = PROP_FLAG_READONLY,
 	},
-	{}
+	{.name = NULL}
 };
 
 /**
@@ -504,7 +514,7 @@ static struct deprecated
 	{"GetTabletID", "TabletID"},
 	{"DebugLevel",	"ToolDebugLevel"},
 	{"CommonDBG",	"TabletDebugLevel"},
-	{"GetTabletID",	"TabletID"},
+	{"Serial",	"BindToSerial"},
 	{"PressCurve",	"PressureCurve"},
 	{"TPCButton",	"TabletPCButton"},
 	{"CursorProx",	"CursorProximity"},
@@ -581,6 +591,8 @@ static struct modifier modifiers[] = {
 	{"lhyper", "Hyper_L"},
 	{"rhyper", "Hyper_R"},
 
+	{"altgr", "ISO_Level3_Shift"},
+
 	{ NULL, NULL }
 };
 
@@ -605,6 +617,8 @@ static struct modifier specialkeys[] = {
 
 	{"backspace", "BackSpace"}, {"Backspace", "BackSpace"},
 
+	{"return", "Return"},
+
 	{"tab", "Tab"},
 
 	{"PgUp", "Prior"}, {"PgDn", "Next"},
@@ -749,8 +763,7 @@ static void usage(void)
 
 static void version(void)
 {
-	printf("%d.%d.%d\n", PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR,
-			     PACKAGE_VERSION_PATCHLEVEL);
+	printf(BUILD_VERSION "\n");
 }
 
 static XDevice* find_device(Display *display, char *name)
@@ -937,6 +950,9 @@ static void list_mod(Display *dpy)
 	m = specialkeys;
 	while(m->name)
 		printf("	%s\n", m++->name);
+
+	printf("Keys not listed above can be specified via their KeySyms."
+		" See the man page for details.\n");
 }
 
 static void list(Display *dpy, int argc, char **argv)
@@ -978,7 +994,7 @@ static const char *convert_specialkey(const char *specialkey)
 			m++;
 	}
 
-	return m->converted ? m->converted : (char*)specialkey;
+	return m->converted ? m->converted : specialkey;
 }
 
 /**
@@ -1005,6 +1021,7 @@ static int special_map_button(Display *dpy, int argc, char **argv, unsigned long
 static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
+static int special_map_panscroll(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 
 /* Valid keywords for the --set ButtonX options */
 static struct keywords {
@@ -1016,6 +1033,7 @@ static struct keywords {
 	{"core", special_map_core},
 	{"modetoggle", special_map_modetoggle},
 	{"displaytoggle", special_map_displaytoggle},
+	{"pan", special_map_panscroll},
 	{ NULL, NULL }
 };
 
@@ -1060,6 +1078,19 @@ static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsign
 	return 0;
 }
 
+static int special_map_panscroll(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size)
+{
+	if (*ndata + 1 > size) {
+		fprintf(stderr, "Insufficient space to store all commands.\n");
+		return 0;
+	}
+	data[*ndata] = AC_PANSCROLL;
+
+	*ndata += 1;
+
+	return 0;
+}
+
 static inline int is_valid_keyword(const char *keyword)
 {
 	struct keywords *kw = keywords;
@@ -1168,9 +1199,12 @@ static int keysym_to_keycode(Display *dpy, KeySym sym)
 	XkbStateRec state;
 	int kc = 0;
 
-
 	if (!xkb)
-		xkb = XkbGetKeyboard(dpy, XkbAllComponentsMask, XkbUseCoreKbd);
+		xkb = XkbGetMap(dpy, XkbAllComponentsMask, XkbUseCoreKbd);
+	if (!xkb) {
+		fprintf(stderr, "Warning: failed to query keyboard map\n");
+		goto out;
+	}
 	XkbGetState(dpy, XkbUseCoreKbd, &state);
 
 	for (kc = xkb->min_key_code; kc <= xkb->max_key_code; kc++)
@@ -2762,6 +2796,24 @@ void argsfromstdin(int *argc, char ***argv)
 }
 #endif /* BUILD_FUZZINTERFACE */
 
+static bool check_for_wayland(Display *dpy)
+{
+	bool		has_xwayland_devices = false;
+	XDeviceInfo	*info;
+	int		ndevices, i;
+
+	info = XListInputDevices(dpy, &ndevices);
+	for (i = 0; i < ndevices; i++) {
+		if (strncmp(info[i].name, "xwayland-", 9) == 0) {
+			has_xwayland_devices = true;
+			break;
+		}
+	}
+	XFreeDeviceList(info);
+
+	return has_xwayland_devices;
+}
+
 int main (int argc, char **argv)
 {
 	int c;
@@ -2847,6 +2899,13 @@ int main (int argc, char **argv)
 		return -1;
 	}
 
+	if (check_for_wayland(dpy)) {
+		fprintf(stderr,
+			"Wayland devices found but this tool is incompatible with Wayland. See\n"
+			"https://github.com/linuxwacom/xf86-input-wacom/wiki/Wayland\n");
+		return 1;
+	}
+
 	if (!do_list && !do_get && !do_set)
 	{
 		if (optind < argc)
@@ -2969,7 +3028,7 @@ static void test_parameter_number(void)
 	 * deprecated them.
 	 * Numbers include trailing NULL entry.
 	 */
-	assert(ARRAY_SIZE(parameters) == 39);
+	assert(ARRAY_SIZE(parameters) == 40);
 	assert(ARRAY_SIZE(deprecated_parameters) == 17);
 }
 
